
#include "local.h"
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "son.h"
#include "timeout.h"

/*************************************************************/
/* Perform an electromagnetic anslysis of microstrip.        */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr21may87 */ 

/* MKS units used at all times */
#include "em.typ"  /* Data structure definitions */

/* insert timing statements by tw 10/08/92 */

/* DOS SECURITY */
double emhl = 0.0;		/* Not used in code */
		      
extern char *getcwd();
char *strcpy();
void exit();
void perror();
#ifndef IBM6000
Rtyp fabs(/* Rtyp x */);
#endif
long init_time=200;
extern char *get_version();
extern int exit_program();
extern int chsm();
sTime *emTime;
extern int chsm_value;
static int check_sum = CHECK_SUM;

int print_license_warning;
char prog_name[] = "em";

main(argc,argv)  int argc; char *argv[];
{ Box_info *Box;        /*Box dimensions, etc.*/
  Cir *cir;      /*Floating pt polygonal circuit descr.*/
  An_info  *An;     /*Analysis control parameters*/
  Fmt_info *Fmt;    /*Name of output format control file*/
  int Ok;               /*Was operation completed properly?*/
  FILE *out;            /*used if end statement needed in response file*/
  Rtyp rtest;
  int i,sum,mem_needed;
  Rtyp ave_eeff();
  int chsm_ret;

  /*Turn on malloc bug detection code*/
  /* Set env. variable DNEW to debug level, typically 2 */
  init_new();
  /*set_new_break_function(809,exit_program);*/

#ifdef EM_TIMING
  W_em_timing(0,NULL);
#endif

  /*Check possible machine dependencies*/
  rtest = 3.9;
  if ( ((int)rtest) != 3 )
  { fprintf(stderr,"\nem: int does not truncate.\n");
    announce_bug("main test");

  }
  if ( check_version_option(argc,argv) ) 
  { print_version(stdout);
    exit(0);
  }

  /* DOS SECURITY */
#ifdef DOS
  verify_hardware_key();
#endif

  /* DOS SECURITY */
  emTime = GetTime( (sTime *)NULL);

  Box = (Box_info *)new(sizeof(Box_info));
  cir = (Cir *)new(sizeof(Cir));
  An = (An_info  *)new(sizeof(An_info));
  init_an(An);
  Fmt = (Fmt_info *)new(sizeof(Fmt_info));

  Ok = Do_command_line( argc,argv,Box,An,Fmt );
  if (Ok<-1) fprintf(stderr,"Command line error.\007\n");
  if (Ok<0) exit(1);

  if (An->State) Title_page();

  /* DOS SECURITY */
  chsm_ret = chsm();

  Ok = Get_circuit( Box->file,Box,cir, PROG_EM, 0 );

  if (Ok==-1) exit(1);
  if (Ok<-1)
  { perror("em");
    if (Ok == -3) fprintf(stderr,"em: unable to open circuit geometry file -- %s --\n",Box->file);
    else fprintf(stderr,"\nem: error reading geometry file.\007\n");
    exit(1); 
  }

  if ( check_cir( Box,cir ) < 0 ) exit(1);

  /* DOS SECURITY */
  init_time -= 30;

  /*If doing both mode removal and de-embedding, issue warning*/
  if (Fmt->dbed && Box->rm_reson==2)
    fprintf(stderr,"em: Warning: With both -B and -d options, bad data may result.\n");

  /*Set output file comment delimiter*/
  switch(Fmt->Cvend)
  { case 'C':
    case 'S': strcpy(Fmt->Com,"*");  break;
    case 'H':
    case 'T': strcpy(Fmt->Com,"!");  break;
    default : Fmt->Com[0] = '\0';
  }

  /* DOS SECURITY */
  /* chsm_ret should be 99 */
  chsm_value = (chsm_ret / 3) - 3;

  Ok = Get_an_param( An );
  if (Ok<-1) fprintf(stderr,"\nem: error reading analysis file.\007\n");
  if (Ok<0) { perror("em"); exit(1); }

  /*Check for error in number of frequencies for SPICE analysis*/
  if (Fmt->spice==1 && (An->Num_freq&1)!=0)
  { fprintf(stderr,"\nem: Lumped spice model must have even number of analysis frequencies.\n");
    exit(1);
  }

  /*If no dbed, set ref to 0 so auto-subsectioning not modifified*/
  if (!Fmt->dbed)
  { Box->rref[0] = 0.0;  Box->ref[0] = 0;
    Box->rref[1] = 0.0;  Box->ref[1] = 0;
    Box->rref[2] = 0.0;  Box->ref[2] = 0;
    Box->rref[3] = 0.0;  Box->ref[3] = 0;
  }

  /*Set up max msub size*/
  if (Box->eeff < 1.0) Box->eeff = ave_eeff(Box->Eps,Box->Mu,Box->Num_levels);
  Box->max_xinc = max_inc(Box->A/Box->Capm,Box->msub_per,Box->eeff,An->fmax);
  Box->max_yinc = max_inc(Box->B/Box->Capn,Box->msub_per,Box->eeff,An->fmax);
  /*Minimum frequency (MHz) below which always use quad precision*/
  Box->qfmin = 3.0e-12 * VEL_LIGHT / sqrt(Box->A * Box->B / (Box->Capm * Box->Capn));

  /*Check if user is using single precision at too low of a freq.*/
  /*If factor of 10.0 is changed, be sure to change similar warning
    messages in two places in emout.c*/
  if (*An->Freq < 10.0 * Box->qfmin && An->Mem_saver)
    { fprintf(stderr,"%s: WARNING: Below %lf MHz use of Memory Saver (-m)",
             prog_name,10.0 * Box->qfmin);
      fprintf(stderr," may result in error.\n\007");
    }

  /*Convert any zero frequencies to an OK value*/
  for (i=0; i<An->Num_freq && *(An->Freq+i) < TINY; i++)
  { *(An->Freq+i) = 0.001 * Box->qfmin;
    fprintf(stderr,"%s: A zero analysis frequency has been specified, using ",
             prog_name);
    fprintf(stderr,"%lf MHz instead.\n",*(An->Freq+i));
  }

  if (An->State)
  { printf("Starting subsectioning...");  fflush(stdout); }
  pcir2cir(cir,Box);
  if (An->State)
    printf("done.\n");

  /*Check subsection size vs. largest frequency*/
  Ok = sub_size(Box->A/Box->Capm,Box->B/Box->Capn,Box->Eps,Box->Mu,cir->n,*(An->Freq+An->Num_freq-1));
  if (Ok<0) exit(1);

  /*assign locations in the system matrix*/
  assign_loc(Box,cir);

  /*Check box parameters for consistency*/
  if (check_box(Box,cir->ports,Fmt->dbed)<0) exit(1);

  /*Transform port indices and set signs*/
  xfrm_ports(cir->ports);
  /*Set up normalizing impedance data*/
  set_port_norm(cir->ports,An);

  /*Initalize Box->met_loss & Box->met_react*/
  ck_surface_imp(cir,Box,An);

  if (An->State)
  { if ( Box->quad_prec )
      printf("Quad precision enabled.\n");
    else if ( *An->Freq < Box->qfmin && sizeof(RRtyp) > sizeof(Rtyp) )
      printf("Quad precision, '-q', automatically enabled below %.3lf MHz.\n",Box->qfmin);
  }

  /*Check to make sure quad provides adavantage*/
  if ( Box->quad_prec && sizeof(RRtyp) <= sizeof(Rtyp) )
    fprintf(stderr,"%s: Warning, the '-q' option has no effect on this computer.\n",prog_name);

  if (An->nsubs_only || (An->jxy == 2 && An->State) )  /*Print out number of subs and quit*/
  { em_mem_calc(cir,Box,An,&mem_needed,&sum);
    printf("%d %d (subsections, Mbytes)\n",sum,mem_needed);
    if ( An->jxy != 2) exit(0);
  }

  if ( An->jxy >= 1 ) Init_jxy_file( argc,argv,Box,cir,An,Fmt );
  if ( An->jxy == 2 ) exit(0);  /*Only subsectioning in jxy file wanted.*/

  Init_rsp_file( argc,argv,Box,cir,An,Fmt );

#ifdef EM_TIMING
  W_em_timing(1,"em Initialization");
#endif

  if (Fmt->dbed) Auto_dbed_loop( Box,cir,An,Fmt );
  else Analysis_loop( Box,cir,An,Fmt );

  /*append an end statement if Cvend format needs it*/
  if (Fmt->Cvend=='H' && Fmt->dbed)
  { out=fopen(Fmt->dbed_file,"a");
    if ( out != NULL ) 
    { fprintf(out,"end\n");
      fclose(out);
    }
  }
  if (Fmt->Cvend=='H' && Fmt->ndbed)
  { out=fopen(Fmt->ndbed_file,"a");
    if ( out != NULL ) {
      fprintf(out,"end\n");
      fclose(out);
    }
  }

  /* DOS SECURITY */
  /* Write out present time to hardware key */
#ifdef DOS
  update_key_end_time();
#endif

  /*Let people know we are done*/
  if (An->State) printf("em: finished.\n\n");

#ifdef EM_TIMING
  W_em_timing(-1,"em Done");
#endif

  exit(0); /* Good finish */
} /*main*/

Title_page()
/*************************************************************/
/* Print out an introduction.                                */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr21may87 */ 
{ 
  printf("\n%62s\n","ELECTROMAGNETIC ANALYSIS OF 3-D PLANAR CIRCUITS");
#ifdef EM_TIMING  
  printf("%35s %s (Clocked)\n\n\n","Version",get_version() );
#else
  printf("%40s %s\n\n","Version",get_version() );
#endif
  printf("%62s\n","(C) Copyright 1987 - 1993 Sonnet Software, Inc.");
  printf("%49s\n\n","All rights reserved.");
} /*Title_page*/

int Do_command_line(argc,argv,Box,An,Fmt)
/*************************************************************/
/* Interpret the command line.                               */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr21may87 */ 
int argc;
char *argv[];
Box_info *Box;
An_info *An;
Fmt_info *Fmt;

{ char *Str;
  int i,bad_file;
  FILE *out;  /*Used to ck if files can be opened*/

  static char def_geo_file[] = DEF_GEO_FILE;
  static char def_an_file[] = DEF_AN_FILE;
  static char def_dbed_file[] = DEF_DBED_FILE;
  static char def_ndbed_file[] = DEF_NDBED_FILE;
  static char def_spice_file[] = DEF_SPICE_FILE;
  static char def_spicet_file[] = DEF_SPICET_FILE;
  static char def_jxy_file[] = DEF_JXY_FILE;
  static char def_cir_name[] = DEF_CIR_NAME;

  init_box(Box); /*Make sure all is cleared out*/
  Box->file = NULL;
  An->file = def_an_file;
  Fmt->dbed_file = NULL;
  Fmt->ndbed_file = NULL;
  Fmt->jxy_file = NULL;
  Fmt->spice_file = NULL;
  Fmt->Param = 's';
  Fmt->ndbed = 0;
  Fmt->dbed  = 0;
  An->Funit = 'X';
  Fmt->Cvend = 'T';
  Fmt->cir_name = def_cir_name;
  Fmt->spice = 0;
  Fmt->spice_prec = 1;
  An->jxy = 0;
  An->Cyc = -1;  /*If left unchanged, signals cmd line did not specify -c# */
  An->Mem_saver = -1;  /*If left unchanged, signals cmd line did not specify -m */
  An->State = 0;
  An->nsubs_only = 0;  /*If set to 1, print out num subs and quit*/
  print_license_warning = -1;	/* by default print license warning */
  An->cans = An->uans = An->jans = NULL;

  if (argc>1 && strcmp(argv[1],"-VV")==0)
  { printf("em: Option no longer enabled.\n");
    exit(0);
  }

  for (i=1; i<argc; i++)
  { Str=argv[i];
    if (*Str=='-') while ( *(++Str)!='\0')  switch ( *Str )  /*Do an option*/
    { case 'j' : An->jxy = 1;  break;  /*Calc. current dist.*/
      case 'J' : An->jxy = 2;  break;  /*Just output subsectioning and quit*/
      case 'c' : An->Cyc = atoi(Str+1);  /*Number of cycles to include*/
		 if (An->Cyc>5) An->Cyc = 5;
		 if (An->Cyc<1) An->Cyc = 1; 
		 while ( isdigit(*(++Str)) );  --Str;   break;
      case 'm' : An->Mem_saver = 1; break;               /*Conserve memory (use single prec for sys matrix)*/
      case 'q' : Box->quad_prec = 1; break;              /*Use long double for matrix fill calculations*/
      case 'z' : Fmt->Param = 'z'; break;                /*Output z parameters*/
      case 'y' : Fmt->Param = 'y'; break;                /*Output y parameters*/
      case 's' : Fmt->Param = 's'; break;                /*Output s parameters*/
      case 'd' : Fmt->dbed  = 1;   break;                /*Output de-embedding parameters*/
      case 'G' : An->Funit  = 'G'; break;                /*GHz for output listing*/
      case 'M' : An->Funit  = 'M'; break;                /*MHz for output listing*/
      case 'H' : Fmt->Cvend = 'H'; break;                /*Generate HP files*/
      case 'S' : Fmt->Cvend = 'S'; break;                /*Generate Midas files*/
      case 'T' : Fmt->Cvend = 'T'; break;                /*Generate Touchstone files*/
      case 'C' : Fmt->Cvend = 'C'; break;                /*Generate Compact files*/
      case 'e' : Box->ckp_lev = 0; break;                /*Don't look up/dwn lev for subsectioning edges*/
      case 'g' : Box->file = ++Str;
		 while( *(++Str)!='\0' ); --Str; break;  /*Circuit geometry data file*/ 
      case 'a' : An->file = ++Str;
		 while( *(++Str)!='\0' ); --Str; break;  /*Analysis control file*/
      case 'r' : if (Fmt->dbed) Fmt->dbed_file = ++Str;  
		 else
		 { Fmt->ndbed_file = ++Str;  Fmt->ndbed = 1; }
		 while( *(++Str)!='\0' ); --Str; break;  /*Output response file*/
      case 'n' : Fmt->cir_name = ++Str;
		 while( *(++Str)!='\0' ); --Str; break;  /*Name for use in cir.rsp file*/
      case 'N' : An->nsubs_only = 1;  break;             /*print num subs only*/
      case 'v' : An->State = 1;  break;  /*Print out current analysis state*/
      case 'Q' : print_license_warning = 0;  break;  	/*Skip printing the license warning message*/
      case 'b' : Box->rm_reson = 1;  break;  /*Just list box resonances*/
      case 'B' : Box->rm_reson = 2;  break;  /*Remove box resonances*/
      case 'x' : Fmt->spice = 1;
                 Fmt->spice_prec = atoi(Str+1);  /*Number of decimal digits*/
                 if ( Fmt->spice_prec < 0 || !isdigit(*(Str+1)) ) Fmt->spice_prec = 1;
		 while ( isdigit(*(++Str)) );  --Str;
                 break;  /*Output spice data*/
      case 'X' : Fmt->spice = 2;
                 break;  /*Output spice data*/
      default  : Usage(1,Str);  return(-1);
    } /*switch*/
    else if ( !strcmp(".geo",&Str[ strlen(Str)-4 ]) )
      Box->file = Str;
    else if ( !strcmp(".an", &Str[ strlen(Str)-3 ]) )
      An->file = Str;
    else if ( !strcmp(".rsp",&Str[ strlen(Str)-4 ]) )
    { if (Fmt->dbed) Fmt->dbed_file = Str;
      else { Fmt->ndbed_file = Str;  Fmt->ndbed = 1; }
    }
    else if ( !strcmp(".d",&Str[ strlen(Str)-2 ]) )
      { Fmt->dbed_file = Str;  Fmt->dbed = 1; }
    else if ( !strcmp(".nd",&Str[ strlen(Str)-3 ]) )
      { Fmt->ndbed_file = Str;  Fmt->ndbed = 1; }
    else if ( !strcmp(".jxy",&Str[ strlen(Str)-4 ]) )
      { Fmt->jxy_file = Str;  An->jxy = 1; }
    else if ( !strcmp(".lc",&Str[ strlen(Str)-3 ]) )
    { Fmt->spice_file = Str;  Fmt->spice = 1; }
    else if ( !strcmp(".lct",&Str[ strlen(Str)-4 ]) )
    { Fmt->spice_file = Str;  Fmt->spice = 2; }

    else if ( Box->file == NULL && i == argc-1 ) /*Last item*/
    { Box->file = cnew( strlen(Str)+5,sizeof(char) );
      sprintf(Box->file,"%s.geo",Str);
      if ( Fmt->dbed && Fmt->dbed_file == NULL )
      { Fmt->dbed_file = cnew( strlen(Str)+3,sizeof(char) );
	sprintf(Fmt->dbed_file,"%s.d",Str);
      }
      if ( !Fmt->dbed && Fmt->ndbed_file == NULL )
      { Fmt->ndbed_file = cnew( strlen(Str)+4,sizeof(char) );
	sprintf(Fmt->ndbed_file,"%s.nd",Str);
      }
      if ( An->jxy && Fmt->jxy_file == NULL )
      { Fmt->jxy_file = cnew( strlen(Str)+5,sizeof(char) );
	sprintf(Fmt->jxy_file,"%s.jxy",Str);
      }
      if ( Fmt->spice == 1 && Fmt->spice_file == NULL )
      { Fmt->spice_file = cnew( strlen(Str)+4,sizeof(char) );
	sprintf(Fmt->spice_file,"%s.lc",Str);
      }
      if ( Fmt->spice == 2 && Fmt->spice_file == NULL )
      { Fmt->spice_file = cnew( strlen(Str)+5,sizeof(char) );
	sprintf(Fmt->spice_file,"%s.lct",Str);
      }
    }

    else { Usage(2,Str); return(-1); }

  } /*for*/

  if ( Fmt->dbed == 0 )
    Fmt->ndbed = 1;  /*Print out non-dbed data at least*/

  if ( Box->file == NULL )
    Box->file = def_geo_file;
  if ( Fmt->dbed_file == NULL && Fmt->dbed )
    Fmt->dbed_file = def_dbed_file;
  if ( Fmt->ndbed_file == NULL && Fmt->ndbed )
    Fmt->ndbed_file = def_ndbed_file;
  if ( Fmt->jxy_file == NULL && An->jxy )
    Fmt->jxy_file = def_jxy_file;
  if ( Fmt->spice_file == NULL && Fmt->spice == 1 )
    Fmt->spice_file = def_spice_file;
  if ( Fmt->spice_file == NULL && Fmt->spice == 2 )
    Fmt->spice_file = def_spicet_file;

  /*Print out results*/
  if (An->State)
  { printf("\n\n\nCircuit geometry file -- %s --\n",Box->file);
    printf("Analysis control file -- %s --\n",An->file);
    if (Fmt->ndbed) printf("Non-de-embedded response file -- %s --\n",Fmt->ndbed_file);
    if (Fmt->dbed) printf("De-embedded response file -- %s --\n",Fmt->dbed_file);
    if (An->jxy==1) printf("Current distribution file -- %s --\n",Fmt->jxy_file);
    if (Fmt->spice==1) printf("Lumped SPICE model file -- %s --\n",Fmt->spice_file);
    if (Fmt->spice==2) printf("Transmission line SPICE file -- %s --\n",Fmt->spice_file);
    if (An->Cyc>0) printf("Using %d cycles in analysis summation.\n",An->Cyc);
  }

  if ( Fmt->dbed && Fmt->ndbed && !strcmp(Fmt->dbed_file,Fmt->ndbed_file) )
  { fprintf(stderr,
      "%s: Cannot store de-embedded and non-de-embedded data in same file.\n",prog_name);
    exit(-1);
  }

  /* DOS SECURITY */
#ifdef DOS
  /* Get time took to get to this point from beginning of program	*/
  /* This will eventually be used in emhh.c to determine if the security*/
  /* check in this program should be done.  It there is too much delay  */
  /* I assumed this is because a hacker is using a debugger and is      */
  /* stepping through the program.  There fore we do not do the security*/
  /* check and just make them fail.  When the hacker gets out of the    */
  /* program it fails for a different reason.				*/
  emTime = GetTime(emTime);

  /* 
  ** Do not check the time if CHECK_SUM is disabled.  The assumption is that
  ** the programmer is debugging the code.
  */
  if ( check_sum == 1 ) {
    emTime->orig_elapse = emTime->elapse;
  }
  
#endif /* DOS */

  /*Check to see that each output file can be opened for append*/
  /*Best to find out now rather than after a few hours of number crunching*/
  bad_file = 0;
  if ( Fmt->ndbed )
  { if ( (out=fopen(Fmt->ndbed_file,"a")) == NULL )
    { fprintf(stderr,"\n%s: Unable to open output data file -- %s --",
				  prog_name,Fmt->ndbed_file);
       bad_file = 1;
    }
    else fclose(out);
  }

  if ( Fmt->dbed )
  { if ( (out=fopen(Fmt->dbed_file,"a")) == NULL )
    { fprintf(stderr,"\n%s: Unable to open output data file -- %s --",
				  prog_name,Fmt->dbed_file);
      bad_file = 1;
    }
    else fclose(out);
  }

  if ( An->jxy==1 ) 
  { if ( (out=fopen(Fmt->jxy_file,"a")) == NULL )
    { fprintf(stderr,"\n%s: Unable to open output data file -- %s --",
				  prog_name,Fmt->jxy_file);
       bad_file = 1;
    }
    else fclose(out);
  }

  if ( Fmt->spice==1) 
  { if ( (out=fopen(Fmt->spice_file,"a")) == NULL )
    { fprintf(stderr,"\n%s: Unable to open output data file -- %s --",
				  prog_name,Fmt->spice_file);
      bad_file = 1;
    }
    else fclose(out);
  }

  if ( Fmt->spice==2 )
  { if ( (out=fopen(Fmt->spice_file,"a")) == NULL )
    { fprintf(stderr,"\n%s: Unable to open output data file -- %s --",
				  prog_name,Fmt->spice_file);
      bad_file = 1;
    }
    else fclose(out);
  }

  if ( bad_file )
  { Str = cnew(500,sizeof(char));
    if ( getcwd(Str,500) != NULL )
      fprintf(stderr,"\n%s: Check write permissions on file name path = %s",
				  prog_name,Str);
    fprintf(stderr,"\007\n\n");
    exit(-1);
  }

  /* DOS SECURITY */
  init_time = emTime->elapse - emTime->orig_elapse;

  return(0);
} /*Do_command_line*/

Usage(Error,Str)  int Error;  char *Str;
/*************************************************************/
/* Print out an error message about command line.            */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr31may87 */ 
{ if (Error==1) fprintf(stderr,"\n\nImproper option - %s -\n",Str);
  if (Error==2) fprintf(stderr,"\n\nImproper argument - %s -\n",Str);
  fprintf(stderr,"Usage -- em [-syzctMGCTHSnrgavdbm] [files]\n\n");
  return;
} /*Usage*/

check_cir(box,cir)  Box_info *box;  Cir *cir;
/*************************************************************/
/* Check for good numbers in cir definition.                 */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr30aug89 */ 
{ Poly_list *plev;
  Rpoly *rpoly;
  Rtyp *pt,*pt_stop,xlim,ylim;
  port_typ *port;
  int ip;

  /*Check box size ratio*/
  if (box->A < 1.0e-10 || box->B < 1.0e-10)
  { fprintf(stderr,"%s: substrate area is too small (or neg.).\n",prog_name);
    return(-1);
  }
  if (box->A > 1.0e10 || box->B > 1.0e10)
  { fprintf(stderr,"%s: substrate area is too big.\n",prog_name);
    return(-1);
  }
  if (box->A/box->B > 199.0 || box->B/box->A >199.0)
  { fprintf(stderr,"%s: substrate area is too narrow.\n",prog_name);
    return(-1);
  }

  /*Check for legal p_ports*/
  if (!cir->ports->n)
  { fprintf(stderr,"%s: there are no ports!\n",prog_name);
    return(-1);
  }
  for (ip=0; ip<cir->ports->n;  ip++)
  { port = cir->ports->port0 + ip;
    if (port->ilev >= cir->n || port->ilev < 0)
    { fprintf(stderr,"%s: port %d on non-existent level.\n",port->index,prog_name);
      return(-1);
    }
    plev = cir->plev0 + port->ilev;
    if (port->poly_index >= plev->n || port->poly_index < 0)
    { fprintf(stderr,"%s: port %d in non-existent polygon.\n",prog_name,port->index);
      return(-1);
    }
    rpoly = plev->rpoly0 + port->poly_index;
    if (port->ipt+1 >= rpoly->n || port->ipt < 0)
    { fprintf(stderr,"%s: port %d on non-existent polygon edge.\n",prog_name,port->index);
      return(-1);
    }
  }

  /* DOS SECURITY */
#ifdef DOS
  /* Basic security check -- actually does not effect the code	*/
  /* This section should be found by the hacker			*/
  { int hladdress,registr,data,status,blkcnt,port;
    char tmpdatastr[9];
    int i;

    for (i=0; i<8; ++i ) tmpdatastr[i] = '0'+i;
    tmpdatastr[0] = '\0';
    port = 0;
    
    hlcrypt(&status,&port,&hladdress,tmpdatastr,&blkcnt);
    status = status *20+3;
    emhl = (double)status + .14159265358979;
  }
#endif /* DOS */

  /*Check for line segments extending outside of box*/
  xlim = 1.00001 * box->A;  ylim = 1.00001 * box->B;
  for (plev=cir->plev0; plev<cir->plev0+cir->n; plev++)
    for (rpoly=plev->rpoly0; rpoly<plev->rpoly0+plev->n; rpoly++)
    { pt_stop = rpoly->xy0 + 2 * rpoly->n;
      for (pt=rpoly->xy0; pt<pt_stop; pt += 2)
        if (*pt<0.0 || *pt>xlim || *(pt+1)<0.0 || *(pt+1)>ylim)
        { xlim = *pt / box->Lcnv;  ylim = *(pt+1) / box->Lcnv;
          fprintf(stderr,"%s: circuit outside of box at (%lf,%lf), level %d.\n",prog_name,xlim,ylim,plev-cir->plev0);
          fprintf(stderr,"%s: all circuit outside of box is ignored.\n",prog_name);
          return(0);
        }
    }
  /*Place no checks following above loop!!!*/

  /* DOS SECURITY */
  emhl /= 2.0;

  return(0);
} /*check_cir*/

check_box(Box,ports,dbed)  Box_info *Box;  port_list *ports;  int dbed;
/*************************************************************/
/* Check for good numbers in Box definition.                 */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr01oct87 */ 
{ int i;
  port_typ *ip,*jp;

  if (Box->quad_prec && sizeof(Rtyp) >= sizeof(RRtyp) )
  { fprintf(stderr,"\n\007%s: WARNING - use of quad precision provides no advantage for\n",prog_name);
    fprintf(stderr,"this version on this computer.\n");
  }

  if (Box->Capm&1 || Box->Capn&1)
  { fprintf(stderr,"\n\007%s: substrate integer dimensions must be even.\n",prog_name);
    return(-1);
  }
  if (Box->Num_levels<1)
  { fprintf(stderr,"\n\007%s: there must be two or more dielectric layers in the box.\n",prog_name);
    return(-1);
  }
  for (i=0; i<=Box->Num_levels; i++)
  { if ( *(Box->H+i)<=0.0 )
    { fprintf(stderr,"\n\007%s: the thickness of layer %d must be greater than zero.\n",prog_name,i);
      return(-1);
    }
    if ( *(Box->H+i) < 5.0e-8 )
    { fprintf(stderr,"\n%s: Warning - the thickness of layer %d is less than 0.05 uM.\n",prog_name,i);
      fprintf(stderr,"%s: Ultra-thin layers may result in numerical precision problems.\n",prog_name);
    }
    if ( *(Box->Eps+i)<TINY )
    { fprintf(stderr,"\n\007%s: dielectric constant of layer %d <= 0.0.\n",prog_name,i);
      return(-1);
    }
    if ( *(Box->Mu+i)<TINY )
    { fprintf(stderr,"\n\007%s: Warning - relative permiability of layer %d <= 0.0.\n",prog_name,i);
      return(-1);
    }
    if ( *(Box->Eps+i)<0.999999*EPSILON_0 )
      fprintf(stderr,"\n\007%s: Warning - dielectric constant of layer %d < 1.0.\n",prog_name,i);
    if ( *(Box->Mu+i)<0.999999*MU_0 )
      fprintf(stderr,"\n\007%s: Warning - relative permiability of layer %d < 1.0.\n",prog_name,i);
  }
  if (Box->A<=0.0 || Box->B<=0.0 || Box->C<=0.0)
  { fprintf(stderr,"\n\007%s: all dimensions of the box must be greater than zero.\n",prog_name);
    return(-1);
  }
  if (Box->Capm<6)
  { fprintf(stderr,"\n\007%s: box must be > 3 cells in x direction.\n",prog_name);
    return(-1);
  }
  if (Box->Capn<6)
  { fprintf(stderr,"\n\007%s: box must be > 3 cells in y direction.\n",prog_name);
    return(-1);
  }

  /*Check for error conditions only on reference planes which include ports*/
  for (ip=ports->port0; ip<ports->port0+ports->n; ip++)
  { if (ip->iside == 0 && Box->ref[0] < 6 && Box->ref[0] > -6 && Box->ref[0] != 0 )
    { fprintf(stderr,"\n\007%s: left ref. plane must be more than 3 cells long.\n",prog_name);
      return(-1);
    }
    if (ip->iside == 1 && Box->ref[1] < 6 && Box->ref[1] > -6 && Box->ref[1] != 0 )
    { fprintf(stderr,"\n\007%s: right ref. plane must be more than 3 cells long.\n",prog_name);
    return(-1);
    }
    if (ip->iside == 2 && Box->ref[2] < 6 && Box->ref[2] > -6 && Box->ref[2] != 0 )
    { fprintf(stderr,"\n\007%s: top ref. plane must be more than 3 cells long.\n",prog_name);
      return(-1);
    }
    if (ip->iside == 3 && Box->ref[3] < 6 && Box->ref[3] > -6 && Box->ref[3] != 0 )
    { fprintf(stderr,"\n\007%s: bottom ref. plane must be more than 3 cells long.\n",prog_name);
      return(-1);
    }
    /*No checks on ref[4] - ref[7], transverse subsection removal plane*/
  }

  /*Check for error conditions on port normalizing impedances*/
  for (ip=ports->port0; ip<ports->port0+ports->n; ip++)
    if ( fabs(ip->rp)+fabs(ip->xp)+fabs(ip->lp)+fabs(ip->cp) < TINY )
    { fprintf(stderr,"\n\007%s: Zero port normalizing impedance not allowed.\n",prog_name);
      return(-1);
    }

  /*Make sure there is at least one port*/
  if (ports->n <= 0)
  { fprintf(stderr,"\n\007%s: geometry file has no ports.\n",prog_name);
    return(-1);
  }
  /*Check if all ports on edge or if any overlap*/
  for (ip=ports->port0; ip<ports->port0+ports->n; ip++)
  { if (ip->x0<0 || ip->x1>Box->Capm || ip->y0<0 || ip->y1>Box->Capn)
    { fprintf(stderr,"%s: port %d is outside of box.\n",prog_name,ip->index);
      return(-1);
    }
    if (ip->iside<0 || ip->iside>5)
    { fprintf(stderr,"%s: warning, port %d has no access to ground.\n",prog_name,ip->index);
    }
    if (dbed && (ip->iside<0 || ip->iside>3) )
    { fprintf(stderr,"%s: port %d is not on box edge.\n",prog_name,ip-ports->port0+1);
      fprintf(stderr,"%s: de-embedding allowed only if all ports on box edge.\n",prog_name);
      return(-1);
    }
    for (jp=ip+1; jp<ports->port0+ports->n; jp++)
    { if ( ip->ilev==jp->ilev && jp->x0<ip->x1 && jp->x1>ip->x0 && jp->y0 == ip->y1)
      { fprintf(stderr,"%s: ports %d and %d overlap.\007\n",ip->index,jp->index,prog_name);
	return(-1);
      }
      if ( ip->ilev==jp->ilev && jp->x0 == ip->x1 && jp->y0<ip->y1 && jp->y1>ip->y0)
      { fprintf(stderr,"%s: ports %d and %d overlap.\007\n",prog_name,ip->index,jp->index);
	return(-1);
      }
    }
    /*Check for a no subsection port*/
    if (ip->nsub <= 0)
    { fprintf(stderr,"\n\007%s: port %d has no subsections.\n",prog_name,ip->index);
      if (Box->Xsym) fprintf(stderr,"%s: if port is below symmetry center line, remove port.\n\n",prog_name);
      return(-1);
    }
    /*Check for a diagonal port*/
    if ( ip->x0 != ip->x1 && ip->y0 != ip->y1 )
    { fprintf(stderr,"\007%s: port %d is diagonal.\n",prog_name,ip->index);
      return(-1);
    }

  }

  return(1);
} /*check_box*/

ck_surface_imp(cir,Box,An) Cir *cir;  Box_info *Box;  An_info *An;
/*************************************************************/
/* Set Box->met_loss and Box->met_react.                     */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr03mar93 */ 
{ int i,loss,react,exists;
  Rtyp *R0,*Rp;
  Rtyp *X0,*Xp;
  Rtyp freq;
  Met_type *Mp;  /*Pointer into Metalization array*/
  Poly_list *plev;
  Rpoly *rpoly;

  loss = react = 0;

  if ( Box->num_met > 0 )
  { /*Calculate each surface resistivity*/
    R0 = (Rtyp *)cnew( Box->num_met,sizeof(Rtyp) );
    X0 = (Rtyp *)cnew( Box->num_met,sizeof(Rtyp) );
    freq = *(An->Freq + An->Num_freq - 1); /*Select highest frequency*/
    for (i=0; i<Box->num_met; i++)
    { Mp = Box->met + i;
      Rp = R0 + i;
      Xp = X0 + i;
      Do_res( Mp->rdc,Mp->rrf,freq,Rp,Xp );
    }

    /*Check for surface resistance*/
    exists = 0;
    for (i=0; i<Box->num_met; i++)
      if ( fabs( *(R0+i) ) > 1.0e-15 )
	exists = 1;

    if (exists)  /*Find out if any non-zero resistance is used*/
      for (plev=cir->plev0; !loss && plev<cir->plev0+cir->n; plev++)
        if ( plev->n > 0 )
          for (rpoly=plev->rpoly0; !loss && rpoly<plev->rpoly0+plev->n; rpoly++)
  	    if ( rpoly->icond >= 0 )
	      if ( fabs( *(R0 + rpoly->icond) ) > 1.0e-15 )
	        loss = 1;

    /*Check for surface reactances*/
    exists = 0;
    for (i=0; i<Box->num_met; i++)
      if ( fabs( *(X0+i) ) > 1.0e-15 )
	exists = 1;

    if (exists)  /*Find out if any non-zero reactance is used*/
      for (plev=cir->plev0; !react && plev<cir->plev0+cir->n; plev++)
        if ( plev->n > 0 )
          for (rpoly=plev->rpoly0; !react && rpoly<plev->rpoly0+plev->n; rpoly++)
  	    if ( rpoly->icond >= 0 )
	      if ( fabs( *(X0 + rpoly->icond) ) > 1.0e-15 )
	        react = 1;

    sFree( (char *)R0 );
    sFree( (char *)X0 );

  }

  Box->met_loss = loss;
  Box->met_react = react;

} /*ck_surface_imp*/

em_mem_calc(cir,Box,An,mbyte_ptr,nsub_ptr) Cir *cir;  Box_info *Box;  An_info *An;
                                           int *mbyte_ptr,*nsub_ptr;
/*************************************************************/
/* Estimate the amount of memory and number of subs required.*/
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr02mar93 */ 
{ Rtyp mem_sys,mem_fft;
  int i,sum;
  msub_list *lev;
  Poly_list *plev;

  /*add up the number of subsections*/
  for (sum=0,i=0; i<=Box->Num_levels; i++)
    sum += (cir->lev0+i)->n;

  mem_sys = 8.0e-6 * sum;
  mem_sys *= sum;  /*so no chance of integer overflow*/

  if ( Box->Lossless && !Box->met_loss )
    mem_sys *= 0.5;

  if (An->Mem_saver)
    mem_sys *= 0.5;

  mem_fft = Box->Capm * Box->Capn * 0.25;  /*FFT matrix*/
  mem_fft += 7.0 * ( Box->Capm + Box->Capn );  /*Est. abt 14 vectors*/
  mem_fft *= 8.0e-6;  /*Convert to Mbytes*/
  if ( !Box->Lossless ) mem_fft *= 2.0;
  if ( Box->quad_prec ) mem_fft *= 2.0;

  *mbyte_ptr = (int)( mem_sys + mem_fft + 1.0 );
  *nsub_ptr = sum;

} /*em_mem_calc*/

Analysis_loop( Box,cir,An,Fmt )  Box_info *Box;  Cir *cir;
			         An_info *An;  Fmt_info *Fmt;
/*************************************************************/
/* Perform a simple analysis (no dbed) over frequency.       */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr09dec87 */ 
{ Rtyp *Fptr;
  int Ok=0;
  complex *Dum1,*Dum2;
  Mat_ptr Cans,Jans;
  sTime *stime;

  /*Initialize timing*/
  stime = GetTime( (sTime *)NULL );

  Dum1 = NULL;  Dum2 = NULL;
  for ( Fptr=An->Freq;  Fptr-An->Freq < An->Num_freq; Fptr++ ) if (Ok==0)
  { An_circuit( Box,cir,An,stime,*Fptr );
    /*Output timing for first frequency only*/
    if ( Fptr == An->Freq )
    { stime = GetTime( stime );
      Out_timing( An,Fmt,stime );  /*sic*/
      sFree( (char *)stime );  stime = NULL;
    }
    shrt_ports(cir->ports,An->cans,An->jans,&Cans,&Jans);
    Mat_close(An->cans);
    An->cans = Cans;
    if (An->jans)
    { Mat_close(An->jans);
      An->jans = Jans;
    }
    if (Box->rm_reson)
    { Out_reson(8,Box->te_res,Box->tm_res,2);
      if (An->State) Out_reson(8,Box->te_res,Box->tm_res,0);
    }
    Ok = Out_result( An,*Fptr,Fmt,Dum1,Dum2 );
    if (An->State) printf("\n");
  }  
  if (Ok<-1) fprintf(stderr,"\nError writing output file.\007\n");
  if (Ok<0) { perror("em");  exit(1); }
} /*Analysis_loop*/

Auto_dbed_loop( Box,cir,An,Fmt )  Box_info *Box;  Cir *cir;
				  An_info *An;  Fmt_info *Fmt;
/*************************************************************/
/* De-embed the discontinuity over frequency.                */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11dec87 */
{ Mat_ptr Rans;  /*Rans stores most recent analysis*/
  Mat_ptr R0,R1; /*R0 is abcd of primary, R1 is dbed abcd*/
  Mat_ptr Port_dbed_mat();
  Mat_ptr Tans,Jans;
  Rtyp *Fptr,Re,Im;
  complex *Ep,*Z0p;
  int i,j,iside,*ilist,num_ports;
  int Ok=0;
  sTime *stime;

  num_ports = cir->ports->n;
  /*Open up space for the Eeff and Z0 arrays*/
  Ep = (complex *)cnew( num_ports,sizeof(complex) );
  Z0p = (complex *)cnew( num_ports,sizeof(complex) );
  /*Set real part of Z0 to -1 as signal not to print*/
  for (i=0; i<num_ports; i++) (Z0p+i)->Re = -1.0;
  /*ilist will contain order of ports in dbed matrix for nodal connection*/
  ilist = (int *)new( 2*num_ports*sizeof(int) );  /*Largest possible size*/
  R0 = Mat_open(2*num_ports,2*num_ports);
  R1 = Mat_open(2*num_ports,2*num_ports);

  /*Initialize timing*/
  stime = GetTime( (sTime *)NULL );

  for ( Fptr=An->Freq;  Fptr-An->Freq < An->Num_freq; Fptr++ ) if (Ok==0)
  { /*Do an analysis of the target discontinuity*/
    if (An->State) printf("Primary structure -- ");
    An_circuit( Box,cir,An,stime,*Fptr );
    if (Box->rm_reson)
    { if ( Fmt->dbed ) Out_reson(8,Box->te_res,Box->tm_res,1);
      if ( Fmt->ndbed ) Out_reson(8,Box->te_res,Box->tm_res,2);
      if ( An->State ) Out_reson(8,Box->te_res,Box->tm_res,0);
    }
    Rans = An->uans = An->cans;  An->cans = NULL;
    /*Fill the primary structure S-parameter matrix*/
    Mat_init(0.0,0.0,R0);
    for (i=0; i<num_ports; i++) for (j=0; j<num_ports; j++)
      Mat_put( i,j, Mat_get_re(i,j,Rans),Mat_get_im(i,j,Rans), R0 );
    /*Connect in the dummy output ports so abcd doesn't blow*/
    for (i=0; i<num_ports; i++)
    { Mat_put_re(i+num_ports,i,1.0,R0);
      Mat_put_re(i,i+num_ports,1.0,R0);
    }

    Rans = NULL;  /*Unde-embedded results saved in An->uans*/
    sabcd(R0);  /*abcd parameters*/

    /*Now go thru each side and get the de-embedding matrix*/
    Mat_init(0.0,0.0,R1);
    for (iside=0; iside<4; iside++)
    { if ( (Rans=Port_dbed_mat(Box,cir,An,*Fptr,iside,ilist,Ep,Z0p))!=NULL)
      { sabcd(Rans);
        for (i=0; i<Rans->Num_col; i++)  for (j=0; j<Rans->Num_col; j++)
          Mat_put( *(ilist+i), *(ilist+j), Mat_get_re(i,j,Rans),Mat_get_im(i,j,Rans), R1);
        Mat_close(Rans);
      }
    }
    
    Mat_mult(R1,R0,&R1);
    abcds(R1);  /*The result is now S-parameters*/

    /*Output timing for first frequency only*/
    if ( Fptr == An->Freq )
    { stime = GetTime( stime );
      Out_timing( An,Fmt,stime );  /*sic*/
      sFree( (char *)stime );  stime = NULL;
    }

    Rans = Mat_open(num_ports,num_ports);
    for (i=0; i<num_ports; i++) for (j=0; j<num_ports; j++)
    { Re = Mat_get_re(i,j,R1);
      Im = Mat_get_im(i,j,R1);
      Mat_put(i,j,Re,Im,Rans);
    }

    /*some ports may be shorted together*/
    shrt_ports(cir->ports,An->uans,An->jans,&Tans,&Jans);
    Mat_close(An->uans);
    An->uans = Tans;
    if (An->jans)
    { Mat_close(An->jans);
      An->jans = Jans;
    }
    shrt_ports(cir->ports,Rans,(Mat_ptr)NULL,&(An->cans),&Jans);
    Mat_close(Rans);  Rans = NULL;
    
    Ok = Out_result(An,*Fptr,Fmt,Ep,Z0p);
    if (Ok<-1) fprintf(stderr,"\nError writing output file.\007\n");
    if (Ok<0) { perror("em");  exit(1); }

    if (An->State) printf("\n");
  }
  Mat_close(R0);  Mat_close(R1);
  sFree( (char *)ilist);
  sFree( (char *)Z0p);  sFree( (char *)Ep);
} /*Auto_dbed_loop*/

Mat_ptr Port_dbed_mat( Box,cir,An,F,iside,ilist,Eeffp,Z0p)  Box_info *Box;
	     Cir *cir;  An_info *An;  Rtyp F;  int iside,*ilist;
	     complex *Eeffp,*Z0p;
/*************************************************************/
/* Create the de-embedding thru and return dbed matrix.      */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11dec87 */
/* jcr25jan88 modified for 2N-port */
/* if no ports on side, return NULL*/
/* The de-embedding matrix is for a given side...
 *      x=0
 * y=0  *---- side 2 ----*
 *      |                |
 *    side 0           side 1
 *      |                |
 *      *---- side 3 ----*
 */
 /*ilist is the list of port's which are de-embedded by Rans*/

{ Rtyp Len0,Lsto,*Tlen,Delta,Eest,Est_eeff();
  static Rtyp dbed_L_sto = -1.0;  /*So we can use the same L for dbed at all freqs*/
  int ILsto,*Tilen;
  Cir *Thru,*make_thru();
  static Cir *old_thru = NULL;
  static Mat_ptr old_result = NULL;
  Mat_ptr Result,dbed_matrix();
  Mat_ptr Rz0;
  int offset,mat_size_sto;
  int sto_sym;
  int sto_tref;  /*Stores transverse sub removal planes*/

  if ( iside<0 || iside>3 )
    announce_bug("Port_dbed_mat(0)");

  /*Store the tr. sub. plane for the side opposite the port*/
  /*and set the side opp. the port to be the same as the port*/
  switch (iside)
  { case 0 : sto_tref = Box->ref[5];
	     Box->ref[5] = Box->ref[4];
	     break;
    case 1 : sto_tref = Box->ref[4];
	     Box->ref[4] = Box->ref[5];
	     break;
    case 2 : sto_tref = Box->ref[7];
	     Box->ref[7] = Box->ref[6];
	     break;
    case 3 : sto_tref = Box->ref[6];
	     Box->ref[6] = Box->ref[7];
	     break;
  }

  if ( iside<2 )  /*If ports are x directed...*/
  { Tlen = &( Box->A );  Tilen = &( Box->Capm ); }
  else
  { Tlen = &( Box->B );  Tilen = &( Box->Capn ); }
  Lsto = *Tlen;  ILsto = *Tilen;
  Delta = Lsto/ILsto;
  mat_size_sto = Box->mat_size;

  offset = Box->ref[iside];
  if ( offset<0 ) offset = -offset; /*if neg, use offset for cal thru length only*/
  if ( offset&1 )
    announce_bug("Port_dbed_mat(1)");
  if (offset==0) /*Select an offset length of about 0.2 wavelengths*/
  { if ( dbed_L_sto <= 0.0 )  /*first time thru only*/
    { Eest = Est_eeff(Box,cir->ports->port0->ilev);
      dbed_L_sto = *Tlen = 0.2 * VEL_LIGHT / (F*1.0E6*sqrt(Eest));
    }
    else *Tlen = dbed_L_sto;
    /*Keep the same delta*/
    *Tilen = *Tlen / Delta;
    if ( (*Tilen &= ~1) > ILsto/2 ) *Tilen = ~1 & ILsto/2;
    if ( *Tilen < 16 ) *Tilen = 16;  /*Set a minimum*/
    *Tlen = *Tilen * Delta;
  }
  else
  { *Tilen = offset;
    *Tlen = Delta * *Tilen;
  }
  if ( Box->ref[iside] < 0 ) offset = 0; /*Signal for zero ref plane*/

  /*Length of first thru for calc of Eeff, in radians*/
  Len0 = *Tlen * TWO_PI * F * 1.0E6 / VEL_LIGHT;

  sto_sym = Box->Xsym;
  if (iside>=2) /*if symmetry on port here, sym makes thru an open*/
    Box->Xsym = 0;

  /*Set up the thru to be analyzed, generate ilist*/
  Thru = make_thru(iside,ilist,cir,Box);

  /*if no ports on this side, return NULL*/
  Result = NULL;
  if (Thru->ports->n > 0) 
  { if (An->State) printf("De-embedding matrix, side %d.\n",iside);
    if ( (iside & 1) && same_cir(Thru,old_thru) )
    /*No need to recalculate if identical to previous thru*/
    { Result = Mat_open(old_result->Num_row,old_result->Num_col);
      Mat_move(old_result,Result);
      dupe_z0(cir->ports,iside,Eeffp,Z0p);
    }
    else
    { assign_loc(Box,Thru);
      xfrm_ports(Thru->ports);  /*Transform port indices and set signs*/
      Rz0 = Mat_open(Thru->ports->n,Thru->ports->n);
      Result = dbed_matrix( Box,Thru,An,F,&Rz0,offset,iside );
      calc_z0(cir->ports,Thru->ports,Rz0,iside,Len0,Eeffp,Z0p);
      Mat_close(Rz0);  Rz0 = NULL;
    }
    /*If symmetry and side 2 port, make debed matrix for two port discontinuities*/
    /*and two t-lines in parallel, do 0.5*B and 2.0*C*/
    if (sto_sym != Box->Xsym) dbed_sym_mod(Result);
  }
  Box->Xsym = sto_sym;

  /*Clear out old result*/
  if (old_thru) { clear_cir(old_thru);  sFree( (char *)old_thru);  old_thru = NULL; }
  if (old_result)  Mat_close(old_result);  old_result = NULL;

  /*Update storage of old result with this result*/
  if ( Result && !(iside & 1) )
  { old_thru = Thru;
    old_result = Mat_open(Result->Num_row,Result->Num_col);
    Mat_move(Result,old_result);
  }

  /*Restore the tr. sub. plane for the side opposite the port*/
  switch (iside)
  { case 0 : Box->ref[5] = sto_tref;
	     break;
    case 1 : Box->ref[4] = sto_tref;
	     break;
    case 2 : Box->ref[7] = sto_tref;
	     break;
    case 3 : Box->ref[6] = sto_tref;
	     break;
  }
  /*Restore other parameters*/
  *Tlen = Lsto;  *Tilen = ILsto;
  Box->mat_size = mat_size_sto;
  return(Result);
} /*Port_dbed_mat*/

dbed_sym_mod(ans) Mat_ptr ans;
/*************************************************************/
/* Modify ABCD in ans for two in parallel.                   */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr01may91 */
{ int i,j,no2;

  no2 = ans->Num_row / 2;
  sabcd(ans);
  for (i=0; i<no2; i++) for (j=0; j<no2; j++)
  { Mat_put(i+no2,j, 2.0*Mat_get_re(i+no2,j,ans), 2.0*Mat_get_im(i+no2,j,ans), ans);
    Mat_put(i,j+no2, 0.5*Mat_get_re(i,j+no2,ans), 0.5*Mat_get_im(i,j+no2,ans), ans);
  }
  abcds(ans);
} /*dbed_sym_mod*/

same_cir(cira,cirb)  Cir *cira,*cirb;
/*************************************************************/
/* If the two circuits are equivalent, return true.          */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28aug89 */ 
/* WARNING -- dielectrics and layer thicknesses NOT checked*/
{ Poly_list *pleva,*plevb;
  int poly_index;
  Ipoly *ipolya,*ipolyb;
  int ip,*ipta,*iptb,*ipt_stop;
  port_typ *porta,*portb;

  if ( !(cira && cirb) ) return(0);
  if (cira->n != cirb->n || cira->ports->n != cirb->ports->n) return(0);

  for (ip=0; ip<cira->ports->n; ip++)
  { porta = cira->ports->port0;  portb = cirb->ports->port0;
    if (porta->x0 != portb->x0 || porta->x1 != portb->x1 ||
        porta->y0 != portb->y0 || porta->y1 != portb->y1 ||
        porta->ilev != portb->ilev) return(0);
  }

  for (pleva=cira->plev0,plevb=cirb->plev0; pleva<cira->plev0+cira->n; pleva++,plevb++)
  { if (pleva->n != plevb->n) return(0);
    for (poly_index=0; poly_index<pleva->n; poly_index++)
    { if ( (pleva->rpoly0+poly_index)->icond != (plevb->rpoly0+poly_index)->icond) return(0);
      ipolya = pleva->ipoly0 + poly_index;  ipolyb = plevb->ipoly0 + poly_index;
      if (ipolya->n != ipolyb->n) return(0);
      ipt_stop = (ipta = ipolya->xy0) +  2 * ipolya->n;
      for (iptb=ipolyb->xy0; ipta<ipt_stop;  ipta++,iptb++)
        if ( *ipta != *iptb ) return(0);
    }
  }

  return(1);
} /*same_cir*/

Rtyp Est_eeff(Box,ilev)  Box_info *Box;  int ilev;
/*************************************************************/
/* Estimate epsilon effective for a t-line.                  */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr14dec87 */ 
{ Rtyp C0,C1,d0,d1;
  int i;

  /*Estimate the epsilon effective by assumming no fringing fields*/
  for (i=0,C0=C1=d0=d1=0.0; i<=Box->Num_levels; i++)
  { if (i<=ilev)
    { C0 += *(Box->H+i) / ( *(Box->Eps+i) * *(Box->Mu+i) );
      d0 += *(Box->H+i);
    }
    else
    { C1 += *(Box->H+i) / ( *(Box->Eps+i) * *(Box->Mu+i) );
      d1 += *(Box->H+i);
    }
  }
  C0 = 1.0/C0;  C1 = 1.0/C1;  d0 = 1.0/d0;  d1 = 1.0/d1;
  return( (C0+C1) / ( (d0+d1)*EPSILON_0*MU_0 ) );
} /*Est_eeff*/

Mat_ptr dbed_matrix( Box,cir,An,F,Rz0p,Offset,iside )
             Box_info *Box;  Cir *cir;  An_info *An;  Rtyp F;
             Mat_ptr *Rz0p;  int Offset,iside;
/*************************************************************/
/* Calculate a single de-embedding matrix.                   */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr09dec87 */ 
{ Mat_ptr R1,R2,Result;
  int xdbed;   /*1 for x directed thru, 0 for y directed thru*/
  int *ilist,*dbed_order();
  Mat_ptr Load_dbed();
  Rtyp Len,derror;

  xdbed = ( cir->ports->port0->iside < 2 );

  /*Do the single length thru*/
  Len = (iside<2) ? Box->A : Box->B;
  Len = Len / Box->Lcnv;
  if (An->State) printf("First standard, %d subs, %lf %s -- ",Box->mat_size,Len,Box->Lstr);
  An_circuit( Box,cir,An,(sTime *)NULL,F );
  R1 = An->cans;  An->cans = NULL;

  if (Box->rm_reson)
  { Out_reson(iside,Box->te_res,Box->tm_res,1);
    if (An->State) Out_reson(iside,Box->te_res,Box->tm_res,0);
  }

  /*Now calculate the double length thru*/
  if (xdbed) X_doub_box(Box,cir);
        else Y_doub_box(Box,cir);
  if (An->State) printf("Second standard, %d subsections -- ",Box->mat_size);
  An_circuit( Box,cir,An,(sTime *)NULL,F );
  R2 = An->cans;  An->cans = NULL;
  if (Box->rm_reson)
  { Out_reson(iside+4,Box->te_res,Box->tm_res,1);
    if (An->State) Out_reson(iside+4,Box->te_res,Box->tm_res,0);
  }

  ilist = dbed_order(cir->ports);
  dbed_new_order(ilist,R1);
  dbed_new_order(ilist,R2);

  Result = Load_dbed(R1,R2,Rz0p,Offset,&derror);

  Out_derror(iside,derror,1);
  if (An->State)
    Out_derror(iside,derror,0);

  abcds(Result);

  dbed_old_order(ilist,Result);

  Mat_close(R1);  Mat_close(R2);
  return(Result);
} /*dbed_matrix*/

calc_z0(cports,tports,Rz0,iside,Len,Eeffp,Z0p)  port_list *cports,*tports;
            int iside;  Mat_ptr Rz0;  Rtyp Len;  complex *Eeffp,*Z0p;
/*************************************************************/
/* Calculate Zo and Eeff (if poss.) from abcd of thru (Rz0). */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr03may91 */ 
{ Mat_ptr T;  /*Merged ports matrix*/
  Mat_ptr dummy;  /*place holder in parameter list*/
  port_typ *iport,*port0,*port_stop;
  int index;
  complex zo,ang,tmp;

  port0 = cports->port0;  /*Primary structure ports*/
  port_stop = port0 + cports->n;

  /*Check for single port on iside*/
  index = -1;
  for (iport=port0; iport<port_stop; iport++) if ( iport->iside == iside )
  { if (index == -1) index = iport->index;
    if (index != iport->index) return;
  }

  /*Check to see that index is not also on other sides*/
  for (iport=port0; iport<port_stop; iport++)
    if ( iport->iside != iside && index == iport->index ) return;

  /*If ports are to be shorted together, do it here for Rz0*/
  if (Rz0->Num_row > 2)
  { abcds(Rz0);
    shrt_ports(tports,Rz0,(Mat_ptr)NULL,&T,&dummy);
    sabcd(T);
  }
  else
  { T =  Mat_open(2,2);
    Mat_move(Rz0,T);
  }

  if (T->Num_row != 2)
    announce_bug("calc_z0");

  /*For two port matrix, get Ang and Z0*/
  zo.Re = Mat_get_re(0,1,T);  zo.Im = Mat_get_im(0,1,T);
  tmp.Re = Mat_get_re(1,0,T);  tmp.Im = Mat_get_im(1,0,T);
  ang.Re = Mat_get_re(0,0,T);  ang.Im = Mat_get_im(0,0,T);  /*Cos(bl)*/
  Mat_close(T);  T = NULL;

  /*skip if cos(bl) close to one*/
  if ( hypot(1.0-ang.Re,ang.Im) < 0.0001 ) return;

  Cdiv(&zo,&tmp,&zo);  Csqrt(&zo,&zo);
  if (zo.Re<0.0) { zo.Re *= -1.0;  zo.Im *= -1.0; } /*Z0 complete*/
  /*Calculate j*sin(bl) */
  Cmult(&zo,&tmp,&tmp);  /*Re(sin(bl)) = Im(temp)*/
  Cacos(&ang,&ang);
  if ( ang.Re<0.0 ) { ang.Re *= -1.0;  ang.Im *= -1.0; }
  if (tmp.Im<0.0)
  { ang.Re = TWO_PI - ang.Re;
    ang.Im *= -1.0;
  }

  /*convert complex ang to Eeff*/
  if ( hypot( ang.Re,ang.Im ) / Len < 0.95  )  ang.Re += TWO_PI;
  ang.Re /= Len;  ang.Im /= Len;
  Cmult( &ang,&ang,&ang );

  (Eeffp + index)->Re = ang.Re;
  (Eeffp + index)->Im = ang.Im;
  (Z0p + index)->Re = zo.Re;
  (Z0p + index)->Im = zo.Im;
} /*calc_z0*/

dupe_z0(ports,iside,Eeffp,Z0p)  port_list *ports;  int iside;  complex *Eeffp,*Z0p;
/*************************************************************/
/* Duplicate Zo and Eeff (if poss.) on iside 2/3 from 0/1.   */
/* (C) COPYRIGHT 1991 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr06may91 */ 
{ port_typ *iport,*jport,*port0,*port_stop;

  if (iside!=1 && iside!=3)
    announce_bug("dupe_z0(0)");

  port0 = ports->port0;
  port_stop = port0 + ports->n;

  jport = port0;
  for (iport=port0; iport<port_stop; iport++) if (iport->iside == iside-1)
  { while(jport->iside!=iside && jport<port_stop) jport++;
    if (jport>=port_stop)
      announce_bug("dupe_z0(1)");
    (Eeffp+jport->index)->Re = (Eeffp+iport->index)->Re;
    (Eeffp+jport->index)->Im = (Eeffp+iport->index)->Im;
    (Z0p+jport->index)->Re = (Z0p+iport->index)->Re;
    (Z0p+jport->index)->Im = (Z0p+iport->index)->Im;
    break;  /*For now only one port possible*/
  }
} /*dupe_z0*/

int *dbed_order(ports)  port_list *ports;
/*************************************************************/
/* List the port numbers for input/output first/last.        */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr27jan88 */ 
/* jcr12oct88 modified for new Cir struct*/
{ int *ilist;    /*List of new port order*/
  int *in,*out;  /*Current position in ilist for input/output ports*/
  int i,xdir;
  port_typ *ip;

  xdir = ( ports->port0->iside <2 );
  ilist = (int *)new( ports->n * sizeof(int) );
  /*get the new order of ports*/
  in = ilist;  out = ilist + ports->n/2;
  for (i=0; i<ports->n; i++)
  { ip = ports->port0+i;
    if ( ((xdir)?ip->x1:ip->y1) <= 1 ) { *in = i;  in++; }
    else { *out = i;  out++; }
  }
  return(ilist);
} /*dbed_order*/
  
dbed_new_order(ilist,M)  int *ilist;  Mat_ptr M;
/*************************************************************/
/* Re-order M according to ilist.                            */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28jan88 */ 
{ Mat_ptr Mt;
  int i,j;
  /*ilist has the new order, copy M to Mt, transfer back to M element
    by element in new order*/
  Mt = Mat_open(M->Num_row,M->Num_col);
  Mat_move(M,Mt);
  for (i=0; i<M->Num_row; i++)  for (j=0; j<M->Num_col; j++)
    Mat_put( *(ilist+i), *(ilist+j), Mat_get_re(i,j,Mt),Mat_get_im(i,j,Mt), M );
  Mat_close(Mt);
} /*dbed_new_order*/

dbed_old_order(ilist,M)  int *ilist;  Mat_ptr M;
/*************************************************************/
/* Return the matrix to it's original order.                 */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28jan88 */ 
{ int i,j;
  Mat_ptr Mt;
  /*ilist has the new order, copy M to Mt, transfer back to M element
    by element in original order*/
  Mt = Mat_open(M->Num_row,M->Num_col);
  Mat_move(M,Mt);
  for (i=0; i<M->Num_row; i++)  for (j=0; j<M->Num_col; j++)
    Mat_put(i,j, Mat_get_re(*(ilist+i),*(ilist+j),Mt),Mat_get_im(*(ilist+i),*(ilist+j),Mt), M );
  Mat_close(Mt);
} /*dbed_old_order*/

Mat_ptr Load_dbed(M1,M2,Mz0p,ref_len,derrorp)  Mat_ptr M1,M2;  Mat_ptr *Mz0p;
                         int ref_len;  Rtyp *derrorp;
/*************************************************************/
/* Load a 2nx2n de-embedding matrix.                         */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr25jan88 */ 
{ Mat_ptr M0;
  int i,j;
  Rtyp derror;

  sabcd(M1);  /*M1_len length thru is abcd parameters*/

  /*Crunch 2*M1_len length thru*/
  sabcd(M2);
  Mat_inv(M2,0);
  M0 = Mat_open(M1->Num_row,M1->Num_col);
  Mat_mult(M1,M2,Mz0p);  /*disc..neg M1 len..neg disc*/
  /*Now get double discontinuity*/
  Mat_mult(*Mz0p,M1,&M0);

  /*Calculate de-embedding error, perfect -> derror = 0.0*/
  /*All diag == 1, all above diag == 0*/
  derror = 0.0;
  for (i=0; i<M0->Num_row; i++)
    derror +=  fabs( 1.0 - Mat_get_mag(i,i,M0) );
  for (i=0; i<M0->Num_col; i++)
    for (j=i+1; j<M0->Num_row; j++)
      derror += Mat_get_mag(i,j,M0);
  *derrorp = 0.01 * derror;

  /*Double port discontinuity in M0, assume no series lumped part...*/
  for (i=0; i<M0->Num_row; i++) for (j=0; j<M0->Num_col; j++)
  { if (i==j) Mat_put(i,j,1.0,0.0,M0);
    else if (i<M0->Num_row/2 || j>M0->Num_col/2)  Mat_put(i,j,0.0,0.0,M0);
    else Mat_put(i,j,0.5*Mat_get_re(i,j,M0),0.5*Mat_get_im(i,j,M0),M0);
  }

  Mat_inv(M0,0);  /*Inverted single discontinuity*/
  Mat_mult(M0,*Mz0p,Mz0p);  /*M1_len de-embeding matrix*/
  Mat_move(*Mz0p,M1);

  /*Calculate an M1_len len line, no disc.*/
  Mat_inv(*Mz0p,0);  /*pos. len line .. disc*/  
  Mat_mult(M0,*Mz0p,Mz0p);  /*pos M1_len line, no disc*/

  if (ref_len!=0) Mat_move(M1,M0);
  return(M0);
} /*Load_dbed*/

Cir *make_thru(iside,ilist,cir,Box)  int iside,*ilist; Cir *cir;
		    Box_info *Box;
/*************************************************************/
/* Create a de-embedding thru for iside.                     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28mar89 */ 
/* jcr28aug89 total re-work, includes cond from rpoly & coplanar waveguide*/
{ Cir *thru;
  port_list *old_ports;
  port_typ *port;
  int i,ip,*ipt,*ipt_stop,ilev;
  int x,y,j,jlim,edge,not_port;
  Poly_list *plev;
  Rpoly *rpoly,*port_poly,*add_rpoly();
  Ipoly *ipoly;
  Rtyp *pt,xyp[10],dx,dy;;

  old_ports = cir->ports;
  thru = (Cir *)new( sizeof(Cir) );
  init_cir( thru,Box->Num_levels );
  /*Copy all old ports on iside to new thru and make the thru*/
  for (ip=0; ip<old_ports->n; ip++) if ( (old_ports->port0 + ip)->iside == iside )
  { *(ilist + thru->ports->n) = ip;
    port = cir->ports->port0 + ip;
    plev = cir->plev0 + port->ilev;
    /*Set up the thru line*/
    port_poly = plev->rpoly0 + port->poly_index;
    pt = port_poly->xy0 + 2 * port->ipt;
    if (iside<2)
    { xyp[0] = 0.0;  xyp[1] = *(pt+1);
      xyp[2] = 0.0;  xyp[3] = *(pt+3);
      /*Make sure of order, so same_cir will work*/
      if (xyp[1]>xyp[3]) { xyp[1] = *(pt+3);  xyp[3] = *(pt+1); }
      xyp[4] = Box->A;  xyp[5] = xyp[3];
      xyp[6] = Box->A;  xyp[7] = xyp[1];
      xyp[8] = 0.0;  xyp[9] = xyp[1];
      edge = port->x0;
    }
    else
    { xyp[0] = *pt;  xyp[1] = 0.0;
      xyp[2] = *(pt+2);  xyp[3] = 0.0;
      if (xyp[0]>xyp[2]) { xyp[0] = *(pt+2);  xyp[2] = *pt; }
      xyp[4] = xyp[2];  xyp[5] = Box->B;
      xyp[6] = xyp[0];  xyp[7] = Box->B;
      xyp[8] = xyp[0];  xyp[9] = 0.0;
      edge = port->y0;
    }
    plev = thru->plev0 + port->ilev;
    rpoly = add_rpoly(plev,xyp,5);
    /*The port sign (+1 or -1) is needed for proper phasing*/
    j = port->sign * (port->index + 1); /*+1 so sign not lost for 0*/
    add_p_port(rpoly-plev->rpoly0,0,port->ilev,j,NULL,thru->ports,0);

    /*Select conductivity for poly*/
    rpoly->icond = port_poly->icond;
    /*Set subsectioning min/max and manhatten flag*/
    rpoly->xmin = port_poly->xmin;
    rpoly->xmax = port_poly->xmax;
    rpoly->ymin = port_poly->ymin;
    rpoly->ymax = port_poly->ymax;
    rpoly->manhatten = port_poly->manhatten;
  }

  /*Update the port coord for following section*/
  port2cir(thru,Box);

  dx = Box->A / Box->Capm;  dy = Box->B / Box->Capn;
  /*Now look for and include any grounded strips on iside*/
  /*No need to ck last lev which is just for gnd vias*/
  for (plev=cir->plev0,ilev=0; ilev<cir->n-1; plev++,ilev++)
    for (i=0; i<plev->n; i++) if ((plev->ipoly0+i)->n>2)
    { port_poly = plev->rpoly0+i;
      ipoly = plev->ipoly0+i;
      ipt_stop = ipoly->xy0 + 2 * ipoly->n - 2;  /*Looking at pairs of pts*/
      for (ipt=ipoly->xy0; ipt<ipt_stop; ipt += 2)
        if ( (iside<2 && *ipt==edge && *(ipt+2)==edge) ||
             (iside>1 && *(ipt+1)==edge && *(ipt+3)==edge) )
        { /*Try to find the edge in the port list*/
          not_port = 1;
          for (ip=0; ip<thru->ports->n && not_port; ip++)
          { port = thru->ports->port0 + ip;
            if (port->ilev == ilev)
            { if (iside<2) not_port = (*(ipt+1)>port->y1 && *(ipt+3)>port->y1) ||
                                      (*(ipt+1)<port->y0 && *(ipt+3)<port->y0);
              else not_port = (*(ipt)>port->x1 && *(ipt+2)>port->x1) ||
                              (*(ipt)<port->x0 && *(ipt+2)<port->x0);
            }
          }
          if (not_port) /*Found it!, add polygon*/
          { if (iside<2)
            { xyp[0] = 0.0;  xyp[1] = dy * *(ipt+1);
              xyp[2] = 0.0;  xyp[3] = dy * *(ipt+3);
              /*Make sure of order, so same_cir will work*/
              if (xyp[1]>xyp[3]) { xyp[1] = dy * *(ipt+3);  xyp[3] = dy * *(ipt+1); }
              xyp[4] = Box->A;  xyp[5] = xyp[3];
              xyp[6] = Box->A;  xyp[7] = xyp[1];
              xyp[8] = 0.0;  xyp[9] = xyp[1];
            }
            else
            { xyp[0] = dx * *ipt;  xyp[1] = 0.0;
              xyp[2] = dx * *(ipt+2);  xyp[3] = 0.0;
              if (xyp[0]>xyp[2]) { xyp[0] = dx * *(ipt+2);  xyp[2] = dx * *ipt; }
              xyp[4] = xyp[2];  xyp[5] = Box->B;
              xyp[6] = xyp[0];  xyp[7] = Box->B;
              xyp[8] = xyp[0];  xyp[9] = 0.0;
            }
            rpoly = add_rpoly(thru->plev0 + ilev,xyp,5);

            /*Select conductivity for poly*/
            rpoly->icond = port_poly->icond;
            /*Set subsectioning min/max*/
            rpoly->xmin = port_poly->xmin;
            rpoly->xmax = port_poly->xmax;
            rpoly->ymin = port_poly->ymin;
            rpoly->ymax = port_poly->ymax;
          }
        }
    }

  /*Finish up ilist*/
  if ( (jlim = thru->ports->n) == 0 ) return(thru);
  for (j=0; j<jlim; j++) *(ilist+jlim+j) = *(ilist+j) + old_ports->n;

  /*Now do the far side ports*/
  for (ip=0; ip<jlim; ip++)
  { port = thru->ports->port0 + ip;
    j = abs(port->index) + cir->ports->n + 1 ;
    if (port->index < 0) j = -j;
    add_p_port(port->poly_index,2,port->ilev,j,NULL,thru->ports,0);
  }

  pcir2cir(thru,Box);

  /*Double the far side in anticipation of 2*L thru, for icond*/
  for (plev=thru->plev0; plev<thru->plev0+thru->n;  plev++)
    for (ipoly=plev->ipoly0; ipoly<plev->ipoly0+plev->n; ipoly++)
    { ipt_stop = ipoly->xy0 + 2 * ipoly->n;
      for (ipt=ipoly->xy0; ipt<ipt_stop; ipt += 2)
        if (iside<2 && *ipt>=Box->Capm) *ipt = 2 * Box->Capm;
        else if (iside>=2 && *(ipt+1)>=Box->Capn) *(ipt+1) = 2 * Box->Capn;
    }

  return(thru);
} /*make_thru*/

X_doub_box( Box,cir )  Box_info *Box;  Cir *cir;
/*************************************************************/
/* Double the box and circuit symetrically in X dimension.   */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr15nov87 */ 
/* jcr06oct88 modified for new Cir struct*/
/* jcr28mar89 modified for new Cir struct*/
{ port_typ *ip;
  msub_list *lev;
  msub *im;
  int i,Capm,Capm2,xy[6];
  
  Capm2 = Box->Capm;
  Capm = Box->Capm *= 2;
  Box->A *= 2.0;

  /*Modify the ports*/
  for (ip=cir->ports->port0; ip<cir->ports->port0+cir->ports->n; ip++) if (ip->iside == 1)
    ip->x0 = ip->x1 = Capm;

  /*Copy all left half subs to right half*/
  for (lev=cir->lev0; lev<cir->lev0+cir->n-1; lev++) for (i=0; i<lev->n; i++)
  { im = lev->msub0+i;  /*msub0 might change during loop*/
    if ( (im->id==MRFX || im->id==MRFY) && *im->ip<Capm2 )
    { xy[0] = Capm - *im->ip;  xy[1] = *(im->ip+1);
      add_msub(im->id,xy,Box,lev);
    }
    if ( (im->id==MXR || im->id==MYR) && *(im->ip+4)<Capm2 )
    { xy[0] = Capm - *im->ip;  xy[1] = *(im->ip+1);
      xy[2] = Capm - *(im->ip+2);  xy[3] = *(im->ip+3);
      xy[4] = Capm - *(im->ip+4);  xy[5] = *(im->ip+5);
      add_msub(im->id,xy,Box,lev);
    }
  }
  assign_loc(Box,cir);
} /*X_doub_box*/

Y_doub_box( Box,cir )  Box_info *Box;  Cir *cir;
/*************************************************************/
/* Double the box and circuit symetrically in Y dimension.   */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr15nov87 */ 
/* jcr06oct88 modified for new Cir struct*/
/* jcr28mar89 modified for new Cir struct*/
{ port_typ *ip;
  msub_list *lev;
  msub *im;
  int i,Capn,Capn2,xy[6];

  Capn2 = Box->Capn;
  Capn = Box->Capn *= 2;
  Box->B *= 2.0;
  
  /*Modify the ports*/
  for (ip=cir->ports->port0; ip<cir->ports->port0+cir->ports->n; ip++) if (ip->iside == 3)
    ip->y0 = ip->y1 = Capn;

  /*Copy all top half subs to bottom half*/
  for (lev=cir->lev0; lev<cir->lev0+cir->n-1; lev++) for (i=0; i<lev->n; i++)
  { im = lev->msub0+i;  /*msub0 might change during loop*/
    if ( (im->id==MRFX || im->id==MRFY) && *(im->ip+1)<Capn2 )
    { xy[0] = *im->ip;  xy[1] = Capn - *(im->ip+1);
      add_msub(im->id,xy,Box,lev);
    }
    if ( (im->id==MXR || im->id==MYR) && *(im->ip+5)<Capn2 )
    { xy[0] = *im->ip;  xy[1] = Capn - *(im->ip+1);
      xy[2] = *(im->ip+2);  xy[3] = Capn - *(im->ip+3);
      xy[4] = *(im->ip+4);  xy[5] = Capn - *(im->ip+5);
      add_msub(im->id,xy,Box,lev);
    }
  }
  assign_loc(Box,cir);
} /*Y_doub_box*/

assign_loc(Box,cir) Box_info *Box;  Cir *cir;
/*************************************************************/
/* Assign system matrix locations and weights.               */
/* (C) COPYRIGHT 1988 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr07oct88 */ 
/* jcr29mar89 modified for new Cir struct*/
{ int i,loc,x,y,max_x,max_y;
  msub *im;
  port_typ *ipt;
  msub_list *lev;
  wt_info **wt0,*wt;
  Rtyp *wtp;
  int cmp_fsub();

  /*Make sure all loc's are -1*/
  for (i=0; i<cir->n; i++)
  { lev = cir->lev0 + i;
    for (im=lev->msub0; im<lev->msub0+lev->n; im++) im->loc = -1;
  }

  /*assign loc for each port subsection*/
  loc = 0;
  for (ipt=cir->ports->port0; ipt<cir->ports->port0+cir->ports->n; ipt++)
  { lev = cir->lev0 + ipt->ilev;
    ipt->nsub = 0;
    for (im=lev->msub0; im<lev->msub0+lev->n; im++)
    { if (im->id<NUM_GSUB) { x = *im->ip;  y = *(im->ip+1); }
      else { x = ( *im->ip + *(im->ip+2) ) >> 1;
             y = ( *(im->ip+1) + *(im->ip+3) ) >> 1;
           }
      if (im->id==MXR) x = *(im->ip+4);
      if (im->id==MYR) y = *(im->ip+5);
      if (ipt->iside<4 && x>=ipt->x0 && x<=ipt->x1 && y>=ipt->y0 && y<=ipt->y1 && im->loc == -1)
      { ipt->nsub++;  im->loc = loc++; }
      /*Catch any via ports*/
      else if (im->id==MVIA && x>=ipt->x0-1 && x<=ipt->x1+1 && y>=ipt->y0-1 && y<=ipt->y1+1 && im->loc == -1)
      { ipt->nsub++;  im->loc = loc++; }
    }
  }

  /*Assign loc for the rest of the msubs*/
  for (lev=cir->lev0; lev<cir->lev0+cir->n; lev++)
    for (im=lev->msub0; im<lev->msub0+lev->n; im++) 
      if (im->loc == -1)
        im->loc = loc++;

   cir->nloc = Box->mat_size = loc;

  /*Assign weight vectors*/
  cir->wt0 = (wt_info **)cnew( NUM_ESUB * loc, sizeof(*cir->wt0) );
  for (lev=cir->lev0; lev<cir->lev0+cir->n; lev++)
    for (im=lev->msub0; im<lev->msub0+lev->n; im++) 
    { wt0 = cir->wt0 + NUM_ESUB * im->loc;
      switch (im->id)
      { case MRFX : wt = *(wt0 + RFX) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+RFX)->weight;
		    /*wy is edge mod*/
		    *wt->wy = (wt->ex0==0 || wt->ex0==Box->Capm) ? 0.5 : 1.0;
		    break;
        case MRFY : wt = *(wt0 + RFY) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+RFY)->weight;
		    /*wy is edge mod*/
		    *wt->wy = (wt->ey0==0 || wt->ey0==Box->Capn) ? 0.5 : 1.0;
		    break;
        case MENDXN :
        case MENDXP :
		    wt = *(wt0 + RFX) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+RFX)->weight;
		    *wt->wy = 0.5;  /*1/2 current for each esub*/
		    
		    wt = *(wt0 + ARFX) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+ARFX)->weight;
		    *wt->wy = (im->id==MENDXN) ? -0.5 : 0.5;  /*1/2 current for each esub*/
		    break;
        case MENDYN :
        case MENDYP :
		    wt = *(wt0 + RFY) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+RFY)->weight;
		    *wt->wy = 0.5;  /*1/2 current for each esub*/
		    
		    wt = *(wt0 + ARFY) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+ARFY)->weight;
		    *wt->wy = (im->id==MENDYN) ? -0.5 : 0.5;  /*1/2 current for each esub*/
		    break;
        case MTPP :
        case MTNP :
        case MTPN :
        case MTNN :
		    wt = *(wt0 + RFX) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = *im->ip - 1;
                    wt->ex1 = *im->ip + 1;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( 2 * sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = *(wt->wx+1) = (Box->sub+RFX)->weight;
		    *wt->wy = (im->id==MTNP || im->id==MTNN) ? -0.5 : 0.5;
		    
		    wt = *(wt0 + RFY) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = *(im->ip+1) - 1;
                    wt->ey1 = *(im->ip+1) + 1;
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( 2 * sizeof(Rtyp) );
		    *wt->wy = *(wt->wy+1) = (Box->sub+RFY)->weight;
		    *wt->wx = (im->id==MTPN || im->id==MTNN) ? -0.5 : 0.5;
		    
		    wt = *(wt0 + LOOP) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+LOOP)->weight;
		    *wt->wy = (im->id==MTPP || im->id==MTNN) ? -0.5 : 0.5;
		    
		    wt = *(wt0 + CROSS) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+CROSS)->weight;
		    *wt->wy = 0.5;  /*1/2 current for each esub*/
		    break;
        case MXR  : wt = *(wt0 + RFX) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = *im->ip + 2;
		    wt->ex1 = *(im->ip+2) - 2;
		    wt->ey0 = *(im->ip+1) + 1;
		    wt->ey1 = *(im->ip+3) - 1;
		    wt->wx = (Rtyp *)new( (wt->ex1 - wt->ex0 + 2) * sizeof(Rtyp) / 2 );
		    wt->wy = (Rtyp *)new( (wt->ey1 - wt->ey0 + 2) * sizeof(Rtyp) / 2 );
		    for (x=wt->ex0,wtp=wt->wx; x<=wt->ex1; x += 2,wtp++)
		    { if ( x <= *(im->ip+4) )
		      { *wtp = x - *im->ip;  *wtp /= *(im->ip+4) - *im->ip; }
		      else
		      { *wtp = *(im->ip+2) - x;  *wtp /= *(im->ip+2) - *(im->ip+4); }
		      *wtp *= (Box->sub+RFX)->weight;
		      if (x==0 || x==Box->Capm) *wtp *= 0.5;  /*Edge mod*/
                    }
		    for (y=wt->ey0,wtp=wt->wy; y<=wt->ey1; y += 2,wtp++)
		    { *wtp = 1.0; }
		    break;
        case MYR  : wt = *(wt0 + RFY) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = *im->ip + 1;
		    wt->ex1 = *(im->ip+2) - 1;
		    wt->ey0 = *(im->ip+1) + 2;
		    wt->ey1 = *(im->ip+3) - 2;
		    wt->wx = (Rtyp *)new( (wt->ex1 - wt->ex0 + 2) * sizeof(Rtyp) / 2 );
		    wt->wy = (Rtyp *)new( (wt->ey1 - wt->ey0 + 2) * sizeof(Rtyp) / 2 );
		    for (x=wt->ex0,wtp=wt->wx; x<=wt->ex1; x += 2,wtp++)
                    { *wtp = 1.0; }
		    for (y=wt->ey0,wtp=wt->wy; y<=wt->ey1; y += 2,wtp++)
		    { if ( y <= *(im->ip+5) )
		      { *wtp = y - *(im->ip+1);  *wtp /= *(im->ip+5) - *(im->ip+1); }
		      else
		      { *wtp = *(im->ip+3) - y;  *wtp /= *(im->ip+3) - *(im->ip+5); }
		      *wtp *= (Box->sub+RFY)->weight;
		      if (y==0 || y==Box->Capn) *wtp *= 0.5;  /*Edge mod*/
                    }
		    break;
        case MVIA : wt = *(wt0 + VIA) = (wt_info *)new( sizeof(wt_info) );
                    wt->ex0 = wt->ex1 = *im->ip;
                    wt->ey0 = wt->ey1 = *(im->ip+1);
		    wt->wx = (Rtyp *)new( sizeof(Rtyp) );
		    wt->wy = (Rtyp *)new( sizeof(Rtyp) );
		    *wt->wx = (Box->sub+VIA)->weight;
		    *wt->wy = 1.0;
		    break;
	default :   announce_bug("assign wt(1)");
      } /*switch*/
    } /*for*/

  /*Find the maximum wy and wx lengths*/
  max_x = 0;  max_y = 0;
  for (wt0=cir->wt0; wt0<cir->wt0+NUM_ESUB*cir->nloc;  wt0++) if (*wt0!=NULL)
  { wt = *wt0;
    if (max_x < wt->ex1 - wt->ex0) max_x = wt->ex1 - wt->ex0;
    if (max_y < wt->ey1 - wt->ey0) max_y = wt->ey1 - wt->ey0;
  }
  cir->max_wt_len = (max_x > max_y) ? max_x + 2 : max_y + 2;

  /*Re-map any subsections past the center line if xsym*/
  if (Box->Xsym)
  { for (wt0 = cir->wt0+RFX; wt0 < cir->wt0+NUM_ESUB*cir->nloc;  wt0 += NUM_ESUB)
      if ((wt = *wt0)!=NULL && wt->ey1>Box->Capn/2)
      { wtp = wt->wy;
        x = (Box->Capn/2 - wt->ey0)>>1;
        y = x + 1;  /*just inside image half*/
	if (2*x+wt->ey0 == Box->Capn/2) x -= 1;  /*image point*/
	max_y = (wt->ey1 - wt->ey0 + 2)>>1;
	while (y<max_y) { *(wtp+x) += *(wtp+y);  x--;  y++; }
	wt->ey1 = Box->Capn/2;
	if ( (wt->ey0&1) != (wt->ey1&1) ) wt->ey1 -= 1;
	if (x < -1)
	  announce_bug("assign wt(2)");
      }
    for (wt0 = cir->wt0+RFY; wt0 < cir->wt0+NUM_ESUB*cir->nloc;  wt0 += NUM_ESUB)
      if ((wt = *wt0)!=NULL && wt->ey1>Box->Capn/2)
      { wtp = wt->wy;
        x = (Box->Capn/2 - wt->ey0)>>1;
        y = x + 1;  /*just inside image half*/
	if (2*x+wt->ey0 == Box->Capn/2) x -= 1;  /*image point*/
	max_y = (wt->ey1 - wt->ey0 + 2)>>1;
	while (y<max_y) { *(wtp+x) -= *(wtp+y);  x--;  y++; }
	wt->ey1 = Box->Capn/2;
	if ( (wt->ey0&1) != (wt->ey1&1) ) wt->ey1 -= 1;
	if (x < -1)
	  announce_bug("assign wt(3)");
      }
  }

} /*assign_loc*/

exit_program(exit_nu)
int exit_nu;
{
	extern void exit();

	exit (exit_nu);
}
